import logging
import pathlib
from uuid import uuid4

import bpy

import blender_asset_tracer.bpathlib

from render_polargrid.validation import is_valid_project_root


logger = logging.getLogger(__name__)


def project_root_changed(self, context):
    # When project root has changed, we need to create a new scene id.
    self._new_scene_id()


class PGScenePropertyGroup(bpy.types.PropertyGroup):
    @classmethod
    def register(cls):
        cls.scene_id = bpy.props.StringProperty(
            name="Scene ID",
            description="Unique ID for this Scene for Polargrid",
            options={'HIDDEN'},
        )

        cls.project_root = bpy.props.StringProperty(
            name="Project Path",
            description=
                "Root path that contains all files related to this project.\n"
                "Use this to select the folder that contains all assets of "
                "the project.\n"
                "When set the file transfer to polargrid will be better "
                "optimized, especially when using linked data.\n"
                "Project Path must be an absolute path",
            subtype='DIR_PATH',
            options=set(),
            update=project_root_changed
        )

    def _ensure_scene_id(self):
        if self.scene_id:
            return

        else:
            self._new_scene_id()

    def _new_scene_id(self):
        self.scene_id = uuid4().hex

    def is_flip_fluids_active(self):
        scene = self.id_data
        if not hasattr(scene, 'flip_fluid'):
            return False

        if not scene.flip_fluid.show_render:
            return False

        for obj in bpy.data.objects:
            if obj.flip_fluid.is_domain():
                return True

        return False

    def get_project_root(self, context):
        default_project_path = self.get_scene_root(context)

        try:
            scene = self.id_data

            pr = self.project_root
            if pr and is_valid_project_root(pr):
                path = blender_asset_tracer.bpathlib.BlendPath(pr.encode('utf-8'))

                if not path.is_absolute():
                    libpath = scene.library.libpath if scene.library else str.encode(str(default_project_path))
                    path = path.absolute(libpath)

                path = pathlib.Path(path.to_path())
                if path.is_dir():
                    return path

        except:
            import traceback; traceback.print_exc()
            logger.error("Exception occured during scene.get_project_root")
            pass

        logger.debug("No project root is set or could be determined. Use current blend file as project root");
        # not usefull setting, return default
        return default_project_path

    def get_scene_root(self, context):
        return pathlib.Path(context.blend_data.filepath).parent
