import pathlib
import logging

import bpy
import asyncio
import threading
import typing

from blender_asset_tracer import pack
from blender_asset_tracer.pack import progress, transfer, PathAction
from blender_asset_tracer.bpathlib import BlendPath

from render_polargrid.polargrid_client import Client


logger = logging.getLogger(__name__)

_running_packer = None  # type: pack.Packer
_packer_lock = threading.RLock()

# For using in other parts of the add-on, so only this file imports BAT.
Aborted = pack.Aborted
FileTransferError = transfer.FileTransferError


class BatProgress(progress.Callback):
    """Report progress of BAT Packing to the UI.

    Uses asyncio.run_coroutine_threadsafe() to ensure the UI is only updated
    from the main thread. This is required since we run the BAT Pack in a
    background thread.
    """

    def __init__(self) -> None:
        super().__init__()
        self.loop = asyncio.get_event_loop()

    def _set_attr(self, attr: str, value):

        async def do_it():
            setattr(bpy.context.window_manager.polargrid, attr, value)

        asyncio.run_coroutine_threadsafe(do_it(), loop=self.loop)

    def _txt(self, msg: str):
        """Set a text in a thread-safe way."""
        logger.info(msg)
        self._set_attr('description', msg)

    def _status(self, status: str):
        """Set the flamenco_status property in a thread-safe way."""
        self._set_attr('status', status)

    def _progress(self, progress: int):
        """Set the flamenco_progress property in a thread-safe way."""
        self._set_attr('progress', progress)

    def pack_start(self) -> None:
        self._txt('Starting BAT Pack operation')
        self._progress(0)

    def pack_done(self,
                  output_blendfile: pathlib.Path,
                  missing_files: typing.Set[pathlib.Path]) -> None:
        if missing_files:
            self._txt('There were %d missing files' % len(missing_files))
        else:
            self._txt('Pack of %s done' % output_blendfile.name)

    def pack_aborted(self):
        self._txt('Aborted')
        self._status('ABORTED')

    def trace_blendfile(self, filename: pathlib.Path) -> None:
        """Called for every blendfile opened when tracing dependencies."""
        self._txt('Inspecting %s' % filename.name)

    def trace_asset(self, filename: pathlib.Path) -> None:
        if filename.stem == '.blend':
            return
        self._txt('Found asset %s' % filename.name)

    def rewrite_blendfile(self, orig_filename: pathlib.Path) -> None:
        self._txt('Rewriting %s' % orig_filename.name)

    def transfer_file(self, src: pathlib.Path, dst: pathlib.Path) -> None:
        self._txt('Uploading %s' % dst.name)

    def transfer_file_skipped(self, src: pathlib.Path, dst: pathlib.Path) -> None:
        self._txt('Skipped %s' % src.name)

    def transfer_progress(self, total_bytes: int, transferred_bytes: int) -> None:
        self._progress(round(100 * transferred_bytes / total_bytes))

    def missing_file(self, filename: pathlib.Path) -> None:
        # TODO(Sybren): report missing files in a nice way
        pass


async def copy(context,
               base_blendfile: pathlib.Path,
               project: pathlib.Path,
               target: pathlib.Path,
               exclusion_filter: str,
               client,
               has_packed_libraries: bool) -> typing.Tuple[pathlib.Path, typing.Set[pathlib.Path], bool]:
    """Use BATðŸ¦‡ to copy the given file and dependencies to the target location.

    :raises: FileTransferError if a file couldn't be transferred.
    :returns: the path of the packed blend file, and a set of missing sources.
    """
    global _running_packer

    loop = asyncio.get_event_loop()

    wm = bpy.context.window_manager

    with PolarGridPacker(bfile=base_blendfile, project=project, target=target, noop=False, client=client) as packer:
        client.modify_scene(scene=context.scene, status='Locked')

        with _packer_lock:
            if exclusion_filter:
                packer.exclude(*exclusion_filter.split())

            packer.progress_cb = BatProgress()
            _running_packer = packer

        logger.debug('awaiting strategise')
        wm.polargrid.status = 'INVESTIGATING'

        if has_packed_libraries:
            await loop.run_in_executor(None, packer.strategise_packed)
        else:
            await loop.run_in_executor(None, packer.strategise_flipfluids)
            await loop.run_in_executor(None, packer.strategise)

        logger.debug('awaiting execute')
        wm.polargrid.status = 'TRANSFERRING'
        await loop.run_in_executor(None, packer.execute)

        logger.debug('done')
        wm.polargrid.status = 'DONE'
        client.modify_scene(scene=context.scene, status='Ok')

    with _packer_lock:
        _running_packer = None

    return packer.output_path, packer.missing_files, packer.use_flip_fluids


def abort() -> None:
    """Abort a running copy() call.

    No-op when there is no running copy(). Can be called from any thread.
    """

    with _packer_lock:
        if _running_packer is None:
            log.debug('No running packer, ignoring call to bat_abort()')
            return
        log.info('Aborting running packer')
        _running_packer.abort()


class PolarGridPacker(pack.Packer):
    def __init__(self, client, **kwargs):
        super().__init__(**kwargs)
        self.client = client
        self._progress_cb = BatProgress()
        self._tscb = pack.progress.ThreadSafeCallback(self._progress_cb)
        self.use_flip_fluids = False

    def _create_file_transferer(self) -> pack.transfer.FileTransferer:
        result = PolarGridTransferrer(self.client)
        result._progress_cb = self._tscb
        return result

    def strategise_flipfluids(self) -> None:
        self.use_flip_fluids = False
        scene = bpy.context.scene
        if not scene.polargrid.is_flip_fluids_active():
            return

        scene_root = bpy.context.scene.polargrid.get_scene_root(bpy.context)
        scene_root = str.encode(str(scene_root))
        for obj in bpy.data.objects:
            if not obj.flip_fluid.is_domain():
                continue

            obj_root = scene_root
            if obj.library is not None:
                obfile = BlendPath(str.encode(obj.library.filepath))
                if obfile.is_blendfile_relative():
                   obfile = obfile.absolute(scene_root)
                obj_root = str.encode(str(obfile.to_path().parent))

            cache_data = obj.flip_fluid.domain.cache

            abs_bpath = BlendPath(str.encode(cache_data.cache_directory)).absolute(obj_root)
            abs_bpath = abs_bpath / str.encode('bakefiles')
            abs_path = pathlib.Path(abs_bpath.to_path())
            if not abs_path.exists():
                continue

            for cache_file in abs_path.iterdir():
                act = self._actions[cache_file]
                asset_pp = self._target_path / cache_file.relative_to(self.project)
                act.new_path = asset_pp
            self.use_flip_fluids = True

    def strategise_packed(self) -> None:
        # NOTE: This needs to be replaced when migrating to next release of BAT.

        # The blendfile that we pack is generally not its own dependency, so
        # we have to explicitly add it to the _packed_paths.
        bfile_path = Path(self.blendfile.absolute())

        # Both paths have to be resolved first, because this also translates
        # network shares mapped to Windows drive letters back to their UNC
        # notation. Only resolving one but not the other (which can happen
        # with the abosolute() call above) can cause errors.
        bfile_pp = self._target_path / bfile_path.resolve().relative_to(self.project.resolve())
        self._output_path = bfile_pp

        act = self._actions[bfile_path]
        act.path_action = PathAction.KEEP_PATH
        act.new_path = bfile_pp


def cleanup_destination_path(dst: pathlib.Path) -> pathlib.Path:
    """
    Version 1.1.1 of BAT adds a drive letter to the dst field of a transfer
    action. This method will remove this as we expect the file to be in the
    root of the project.
    """
    if dst.drive:
        dst = dst.relative_to(*dst.parts[:1])
    if dst.suffix == ".polar":
        dst = dst.with_suffix(".blend")
    return dst


class PolarGridTransferrer(pack.transfer.FileTransferer):
    def __init__(self, client):
        super().__init__()
        self.client = client

    def run(self) -> None:
        # Retrieve all known server hashes from the server.
        server_hashes = self.client.init_upload_files()

        for src, dst, act in self.iter_queue():
            dst = cleanup_destination_path(dst)
            logger.info("check {{ src={}, dst={}, act={} }}".format(src, dst, act))
            if src.is_file():
                self.client.upload_file((src, dst), server_hashes, callback=self.report_transferred)

                # Remove temp files created during the BAT process
                if act == pack.transfer.Action.MOVE:
                    logger.debug("remove temp file {{ src={} }}".format(src))
                    src.unlink()

            # Dir upload doesn't seem to be needed by BAT.
            # elif src.is_dir():
                # pass
